#!/usr/bin/env python

""" Main entry of the script.
"""

import argparse
import sys
import os
import gzip
import datetime
import numpy as np
from operator import itemgetter, attrgetter

# import internal library
from . import kitsunejf as jf

USAGE = """
"""

def main():
    commands = {
                "cre": cumulative_relative_entropy,
                "acf": average_common_feature,
                "ofc": observe_feature_occurrence,
                "dmatrix": generate_distance_matrix
               }

    parser = argparse.ArgumentParser(description="Compute k-mer signature of genome",
                                     usage="""kitsune <command> [<args>]

Commands can be:
cre <filename>                    Compute cumulative relative entropy.
acf <filenames>                   Compute average number of common feature between signatures.
ofc <filenames>                   Compute observed feature frequencies.
dmatrix <filenames>               Compute distance matrix.
""")
    parser.add_argument('command')
    args = parser.parse_args(sys.argv[1:2])
    if args.command not in commands:
        parser.print_help()
        sys.exit(1)

    cmd = commands.get(args.command)
    cmd(sys.argv[2:])

def cumulative_relative_entropy(args):
    from . import cre
    """ Calculate optimal k-mer through CRE value.

    Args:
        args (TODO): TODO

    Returns: TODO

    """
    desc = "Calculate k-mer from cumulative relative entropy of all genomes"
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument("filename", type=str, help="a genome file in fasta format")
    parser.add_argument("--fast", action="store_true", help="Jellyfish one-pass calculation (faster)")
    parser.add_argument("--canonical", action="store_true", help="Jellyfish count only canonical mer (use for raw read count)")
    parser.add_argument("-ke", "--kend", required=True, type=int, help="last k-mer")
    parser.add_argument("-kf", "--kfrom", default=4, type=int, help="Calculate from k-mer")
    parser.add_argument("-o", "--output", type=str, help="output filename")
    args = parser.parse_args(args)

    outdata = cre.cal_cre(args.filename, **vars(args))
    outdata = sorted(outdata.items(), key=itemgetter(0))
    outdata = '\n'.join(['\t'.join([str(x) for x in data]) for data in outdata])
    if args.output is not None:
        with open(args.output, 'w') as ofhandle:
            ofhandle.write(outdata)
    else:
        print(outdata)


def average_common_feature(args):
    from . import acf
    """ Calculate an average number of common feature pairwise
        between one genome against others

    Args:
        args (TODO): TODO

    Returns: TODO

    """
    desc = "Calculate average number of common feature"
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument("filenames", nargs="+", help="genome files in fasta format")
    parser.add_argument("--fast", action="store_true", help="Jellyfish one-pass calculation (faster)")
    parser.add_argument("--canonical", action="store_true", help="Jellyfish count only canonical mer (use for raw read count)")
    parser.add_argument("-k", "--kmers", nargs="+", required=True, type=int)
    parser.add_argument("-o", "--output", type=str, help="output filename")
    args = parser.parse_args(args)

    outdata = cal_acf(args.filenames, **vars(args))
    outdata = sorted(outdata.items(), key=itemgetter(0))
    outdata = '\n'.join(['\t'.join([str(x) for x in data]) for data in outdata])
    if args.output is not None:
        with open(args.output, 'w') as ofhandle:
            ofhandle.write(outdata)
    else:
        print(outdata)

def observe_feature_occurrence(args):
    from . import ofc
    """ Calculate an observe feature frequency

    Args:
        args (TODO): TODO

    Returns: TODO

    """

    desc = "Calculate observe feature occurrence"
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument("filenames", nargs="+", help="genome files in fasta format")
    parser.add_argument("--fast", action="store_true", help="Jellyfish one-pass calculation (faster)")
    parser.add_argument("--canonical", action="store_true", help="Jellyfish count only canonical mer (use for raw read count)")
    parser.add_argument("-k", "--kmers", nargs="+", required=True, type=int)
    parser.add_argument("-o", "--output", type=str, help="output filename")
    args = parser.parse_args(args)

    outdata = cal_ofc(args.filenames, **vars(args))
    outdata = sorted(outdata.items(), key=itemgetter(0))
    outdata = '\n'.join(['\t'.join([str(x) for x in data]) for data in outdata])
    if args.output is not None:
        with open(args.output, 'w') as ofhandle:
            ofhandle.write(outdata)
    else:
        print(outdata)

def generate_distance_matrix(args):
    from . import matrix
    """Generate distance matrix base on k-mer

    The output will

    Args:
        args (TODO): TODO

    Returns: TODO

    """
    desc = "Calculate a distance matrix"
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument("filenames", nargs="+", help="genome files in fasta format")
    parser.add_argument("--fast", action="store_true", help="Jellyfish one-pass calculation (faster)")
    parser.add_argument("--canonical", action="store_true", help="Jellyfish count only canonical mer (use for raw read count)")
    parser.add_argument("-k", "--kmer", required=True, type=int)
    parser.add_argument("-o", "--output", type=str, help="output filename")
    parser.add_argument("-t", "--thread", type=int, default=1)
    parser.add_argument("-d", "--distance", default="cosine")
    parser.add_argument("-f", "--format", default="phylip")
    args = parser.parse_args(args)
    args.dist_func = jf.DISTANCE_FUNCTION[args.distance]
    outdata = matrix.cal_matrix(args.filenames, **vars(args))
    if args.format == "phylip":
        outdata = str(len(outdata)) + '\n' + '\n'.join([k + '\t' + '\t'.join([str(x) for x in v]) for k, v in outdata.items()])
    if args.output is not None:
        with open(args.output, 'w') as ofhandle:
            ofhandle.write(outdata)
    else:
        print(outdata)

if __name__ == "__main__":
    main()



from . import kitsunejf as jf
import math
from tqdm import tqdm

def cal_cre(fsa, kend, kfrom=4, **karg):
    """ Calculate Cumulative Relative Entropy (CRE)
        CRE = sum(RE from kmer to infinite)
    Args:
        fsa genome file in fasta format
        kend the infinite kmer
        kfrom calculate to (defualt=4)

    Returns: dict(kmer: CRE)

    """
    a0 = None
    a1 = None
    a2 = None
    result = {}
    for kmer in tqdm(range(kend, kfrom-1, -1)):
        if a0 is None:
            a0 = jf.Kmercount(fsa, kmer, **karg)
            a1 = jf.Kmercount(fsa, kmer - 1, **karg)
            a2 = jf.Kmercount(fsa, kmer - 2, **karg)
        else:
            a0 = a1
            a1 = a2
            a2 = jf.Kmercount(fsa, kmer - 2, **karg)
        if kmer + 1 in result:
            result[kmer] = cal_re(a0, a1, a2) + result[kmer + 1]
        else:
            result[kmer] = cal_re(a0, a1, a2)
    return result

def cal_re(a0, a1, a2):
    """ Calculate Relative Entropy (RE)
        f' = fl * fr / fb
        example: f' of mer 'ATTTGCA' 
                f  is frequency of ATTTGCA from a0
                fl is frequency of ATTTGC- from a1
                fr is frequency of -TTTGCA from a1
                fb is frequency of -TTTGC- from a2
        RE = f * log2(f/f')
    Args:
        a0 Kmercount kmer
        a1 Kmercount kmer - 1
        a2 Kmercount kmer - 2

    Returns: float

    """
    result = 0
    rfactor = a0.sum
    lfactor = math.log((a1.sum ** 2) / (a2.sum * rfactor), 2)
    for key in a0.keys():
        realf = a0[key]
        left = a1[key[0:-1]]
        right = a1[key[1:]]
        below = a2[key[1:-1]]
        if 0 not in (left, right, below):
            expectf = left * right / below
            result += max(0, realf / rfactor * (math.log(realf / expectf, 2) + lfactor))
    return result

if __name__ == "__main__":
    print(cal_cre('./examples/ASM170810v1_genomic.fna', 25))


import os
import sys
import copy
import shutil
import platform
import tempfile
import subprocess
import collections
import numpy as np
from tqdm import tqdm
from scipy.spatial import distance

__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))

class NoInput(Exception):
    """Do X and return a list."""
    pass

if platform.system() == 'Darwin':
    jellyfishpath = os.path.join(__location__, 'modules', "jellyfish-macosx")
elif platform.system() == 'Linux':
    jellyfishpath = os.path.join(__location__, 'modules', "jellyfish-linux")
else:
    raise

DISTANCE_FUNCTION = {
    'braycurtis': distance.braycurtis,
    'canberra': distance.canberra,
    'chebyshev': distance.chebyshev,
    'cityblock': distance.cityblock,
    'correlation': distance.correlation,
    'cosine': distance.cosine,
    'dice': distance.dice,
    'euclidean': distance.euclidean,
    'hamming': distance.hamming,
    'jaccard': distance.jaccard,
    'kulsinski': distance.kulsinski,
    'matching': distance.matching,
    'rogerstanimoto': distance.rogerstanimoto,
    'russellrao': distance.russellrao,
    'sokalmichener': distance.sokalmichener,
    'sokalsneath': distance.sokalsneath,
    'sqeuclidean': distance.sqeuclidean,
    'yule': distance.yule
}

class Kmercount(collections.Counter):
    """Do X and return a list."""
    def __init__(self, fsa, kmer = 21, **karg):
        if 'thread' not in karg:
            karg['thread'] = 1
        if 'lower' not in karg:
            karg['lower'] = 1
        if 'bchashsize' not in karg: #hashsize for jellyfish bc step
            karg['bchashsize'] = '1G'
        if 'hashsize' not in karg: #hashsize for jellyfish count step
            karg['hashsize'] = '100M'
        if 'canonical' not in karg:
            canonical = ''
        else:
            canonical = '-C'


        if not os.path.isfile(fsa):
            raise NoInput('input is missing')

        filebasename = os.path.basename(fsa)

        with tempfile.TemporaryDirectory() as tmpdirname:

            if 'fast' in karg and karg['fast']:
                # for genome with one step
                dumpdata = subprocess.getoutput("""
                    {0} count {8} -m {1} -s {3} -t {4} {6} -o {5}.jf
                    {0} dump -c -L {7} {5}.jf
                    """.format(
                        jellyfishpath,
                        kmer,
                        karg['bchashsize'],
                        karg['hashsize'],
                        karg['thread'],
                        os.path.join(tmpdirname, filebasename),
                        fsa,
                        karg['lower'],
                        canonical
                    )
                )
            else:
                dumpdata = subprocess.getoutput("""
                    {0} bc {8} -m {1} -s {2} -t {4} -o {5}.bc {6}
                    {0} count {8} -m {1} -s {3} -t {4} --bc {5}.bc {6} -o {5}.jf
                    {0} dump -c -L {7} {5}.jf
                    """.format(
                        jellyfishpath,
                        kmer,
                        karg['bchashsize'],
                        karg['hashsize'],
                        karg['thread'],
                        os.path.join(tmpdirname, filebasename),
                        fsa,
                        karg['lower'],
                        canonical
                    )
                )

        datadict = {}
        for line in dumpdata.split('\n'):
            dat = line.rstrip().split(' ')
            datadict[dat[0]] = int(dat[1])

        super(Kmercount, self).__init__(datadict)

        # assign instance variable
        self.sum = sum(self.values())
        self.name = '.'.join(filebasename.split('.')[0:-1]).replace(' ', '_')

    def __repr__(self):
        return self.name

    def dist(self, other, dist_func = distance.cosine):
        """Do X and return a list."""
        a, b = self.norm(other)
        return dist_func(a, b)

    def norm(self, other):
        """Do X and return a list."""
        mers = list(self.keys())
        mers.extend(list(other.keys()))
        mers = list(set(mers))
        a = []
        b = []
        for mer in mers:
            a.append(self[mer])
            b.append(other[mer])
        return np.array(a), np.array(b)

if __name__ == "__main__":
    genomea = Kmercount('../examples/S288C_reference_sequence_R64-2-1_20150113.fsa')
    genomeb = Kmercount('../examples/ASM170810v1_genomic.fna')
    print(genomea.dist(genomeb))


from . import kitsunejf as jf
import math
import collections
import numpy as np
from tqdm import tqdm

def cal_ofc(fsas, kmers, **karg):
    """ Calculate shannon entropy of observed feature occurrences (OFC)
        ofc(l) = -sum(p ln p)
    Args:
        fsa a genome file
        kmers a list of kmer to calculate

    Returns: float shannon entropy

    """
    result = {}
    
    for kmer in tqdm(kmers):
        keys = []
        for fsa in fsas:
            keys.extend(list(jf.Kmercount(fsa, kmer, **karg).keys()))
        
        count_feature = list(collections.Counter((collections.Counter(keys).values())).values())
        lnp = np.log2(count_feature) - (kmer * 2)
        result[kmer] = np.sum(np.exp2(lnp) * lnp) * -1

    return result

if __name__ == "__main__":
    print(cal_ofc(['./examples/ASM170810v1_genomic.fna', './examples/ASM170810v1_genomic.fna'], [10,11,12]))



from . import kitsunejf as jf

import math
import collections
import numpy as np
from tqdm import tqdm
from operator import attrgetter

def cal_matrix(fsas, kmer, dist_func=jf.DISTANCE_FUNCTION['cosine'], **karg):
    result = collections.OrderedDict()
    counters = []
    n = len(fsas)
    # create kmercounter
    for fsa in tqdm(fsas):
        counters.append(jf.Kmercount(fsa, **karg))

    sorted(counters, key=attrgetter('name'))

    for counter in counters:
        result[counter.name] = [0] * n

    for pri_idx, pri_counter in enumerate(counters):
        for sec_idx in range(pri_idx + 1, n):
            sec_counter = counters[sec_idx]
            dist = pri_counter.dist(sec_counter, dist_func)
            result[pri_counter.name][sec_idx] = dist
            result[sec_counter.name][pri_idx] = dist
    return result

if __name__ == "__main__":
    print(cal_matrix(['./examples/S288C_reference_sequence_R64-2-1_20150113.fsa', './examples/ASM170810v1_genomic.fna'], 25))



from . import kitsunejf as jf
import math
from tqdm import tqdm

def cal_acf(fsas, kmers, **karg):
    """ Calculate Average number of common features (ACF)
        acf(l) = sum((common feature of genome i and j at length l)/number of genome - 1)
    Args:
        fsas a list of genome file
        kmers a list of kmer to calculate

    Returns: dict(kmer: acf)

    """
    n = len(fsas)
    if n >= 2:
        result = {}
        for kmer in tqdm(kmers):
            keys_array = []
            for fsa in fsas:
                keys_array.append(set(jf.Kmercount(fsa, kmer, **karg).keys()))
            ccf = 0
            for pri_idx, key in enumerate(keys_array):
                for sec_idx in range(pri_idx + 1, n):
                    ccf += len(keys_array[pri_idx] & keys_array[sec_idx])
            result[kmer] = ccf/(n-1)
        return result
    else:
        raise

if __name__ == "__main__":
    print(cal_acf(['./examples/ASM170810v1_genomic.fna', './examples/ASM170810v1_genomic.fna'], [10,11]))



# -*- coding: utf-8 -*-

##### Setup
from setuptools import setup, find_packages

def readme():
    with open('README.rst') as f:
        return f.read()

setup(
    name='kitsune',
    version='1.0.0',
    description='tools for finding an optimal kmer',
    long_description=readme(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
    ],
    keywords='kitsune',
    url='https://gitlab.com/natapol.por/ksiga',
    packages=['kitsune'],
    author='Natapol Pornputtapong',
    author_email='natapol.por@gmail.com',
    license="Apache 2.0",
    scripts=[],
    entry_points = {
        'console_scripts': ['kitsune=kitsune.main:main'],
    },
    include_package_data=True,
    install_requires=["numpy >= 1.1.0",
                        "scipy>=0.18.1",
                        "biopython>=1.68",
                        "tqdm"
                    ],
    zip_safe=False
)
